
#include "stdafx.h"
#include "stackTraceC.h"
#include "cubeData.hpp"
#include "sep_params.h"
#include "misc.h"
#include "mpi_util.h"

#include "my_timer.h"
//#include <complex>
#include <fftw3.h>
#include <boost/shared_array.hpp>
#include <boost/shared_ptr.hpp>

#define PLAN_FLAGS FFTW_PATIENT
//FFTW_EXHAUSTIVE FFTW_PATIENT

#define exeFFTW(x) {VERIFY(NULL != (x));}

//Make sure the input data is allocated on aligned memory address.
/**
 * Test FFTW program, these are three programs to be considered:
 * 1. Transpose and do FFTW
 *    1. Transpose using MKL
 *    2. Transpose use dummy code
 * 2. FFTW on slow axis
 * 3. MKL on slow axis
 * 4. vectorized FFTW
 */


using namespace Yang;
using namespace std;

//Generate vec from 0.0-1.0, uniform distribution
void random_vec (float* x, long long n) {
  VERIFY(RAND_MAX > 32000);
  for (long long i=0; i<n; i++) {
    x[i] = float(rand() % RAND_MAX) / float(RAND_MAX) ;
  }
}

typedef enum __FTtype {
  FFTW2_FPU,
  TRANSP_FFTW1_FPU,
  FFTW2_SSE,
  FFTW2_FPU_MULTH, //multi-thread FFTW, requires libfftw3l_threads.a library, which is not built in this contex.
  FFTW2_FPU2
} FTtype; //different approaches

#define fftwf_alloc_complex(n) ((fftwf_complex* )fftwf_malloc(sizeof(fftwf_complex)*n))

int main(int argc, char *argv[]) {
  initCallStackTrace();
  //have to init par
  sep_params *pars = new sep_params(argc, argv);

  int thrSave = pars->get_int("nth_save",1);
  //Thread to save
  PRV(thrSave); init_omp(thrSave);
  cubeDataF inCube; //might need to ask cube to align the address to 16 bytes.
  //define the problem size
  bool bHasInput = tagExist("input");
  PRV(bHasInput);
  int n1,n2;
  if (bHasInput) { //read input, otherwise
    //Load the data
    inCube.load("input",2);
  } else {
    n1 = pars->get_int("n1");
    n2 = pars->get_int("n2");
    hypercube shp(n1,n2);
    inCube.alloc(shp);
    //and make up random numbers on it
    random_vec(inCube.dataPtr(),shp.get_n123());
  }
  
  int nth = g_NUM_THREADS; PRV(g_NUM_THREADS);
  const hypercube& shp = inCube.domain();
  n1 = shp.get_n(1); n2 = shp.get_n(2);
  
  vecAxis axes = shp.get_axes();
  //VERIFY(axes[1].n%64==0);
  get_axis_inv(axes[1], axes[1].n, axes[1]); 
  hypercube shapeOut; shapeOut.set_axes(axes);

  //Do the fft, might need to align the data.
  hypercube shpOutcube(shp);
  
  long n123 = shapeOut.get_n123();  long n12 = shapeOut.get_n23(1, 2);
  bool bInPlace = pars->get_bool("bInPlace", true); PRV(bInPlace);
  bool bMakesureAligned = pars->get_bool("bAligned",true);
  
  //create the fftw plan
  fftwf_complex *fftIn,*fftOut;
  fftwf_complex *fftwin[nth],*fftwinOut[nth];
  fftwf_plan FFTplan[nth];
  int ith;
  
  PRV(n12);
  //allocate the aligned memory
  fftIn   = (fftwf_complex * )fftwf_malloc(sizeof(fftwf_complex) * n12);
  memset(fftIn, 0, n12 * sizeof(fftwf_complex)); 
  if (!bInPlace) {
    fftOut  = (fftwf_complex * )fftwf_malloc(sizeof(fftwf_complex) * n12);
  } else {
    fftOut = fftIn;
  }

  //different approach, measure the time consumed.
  pair<int, FTtype> MT_match_table[] = {make_pair(0, FFTW2_FPU), make_pair(1, TRANSP_FFTW1_FPU),
    make_pair(2, FFTW2_SSE), make_pair(3,FFTW2_FPU_MULTH),make_pair(4,FFTW2_FPU2)};

  FTtype method = findEnumItemFromPar(pars, MT_match_table, NELEM(MT_match_table), "method", 0);
  
  int rank = 1; int no[3] = { 1, axes[0].n, axes[1].n };  PRVEC(no, 3);
  float scalar = 1.0f;
  int dir = (true) ? FFTW_FORWARD : FFTW_BACKWARD;
  Yang::timer clk;
  int numTransforms[nth],oTransforms[nth];
  //copy the data in
  //std::copy(inCube.dataPtr(),inCube.dataPtr()+n12,fftIn); 

  float totRunTime = 0.0f;
  //Repeat several times for better measurement
  for (int iter=0; iter<10; iter++) {
    for (long i = 0; i < n12; i++) {
      fftIn[i][0] = inCube[i];
      fftIn[i][1] = 0.0f;
    }
    if (method==FFTW2_FPU) {
      //PRV2(fftIn,fftIn+2);  PRV2(&fftIn[0],&fftIn[2]);
      //Now the pointers for each thread will point to a different locations.
      //make sure the address is aligned.
      if (bMakesureAligned) {
        //According to nth, align the memory
        int NS = n1/16; //total number fo segments
        VERIFY(n1%16==0);
        //split the job
        workScheduler1d wrk(NS, nth, Yang::workScheduler1d::chunk);
        int n,o,d;
        for (ith = 0; ith < nth; ith++) {
          wrk.get_nod(ith,&n,&o,&d);
          fftwin[ith]     = &fftIn[o*16];
          fftwinOut[ith]  = &fftOut[o*16]; 
          numTransforms[ith] = n*16;
          oTransforms[ith] = o*16;
          //TRACE(("ith=%d, %d,%d,%d\n",ith,n,o,d));
        }
      } else { //not necessarily aligned.
        //split the job
        int NS = n1;
        workScheduler1d wrk(NS, nth, Yang::workScheduler1d::chunk);
        int n, o, d;
        for (ith = 0; ith < nth; ith++) {
          wrk.get_nod(ith, &n, &o, &d);
          fftwin[ith]         = &fftIn[o];
          fftwinOut[ith]      = &fftOut[o];
          numTransforms[ith]  = n;
          oTransforms[ith]    = o;

          if (iter==0) {
            TRACE(("ith=%d, %d,%d,%d\n",ith,n,o,d));
          }
        }
      }
      
      //clk.reset("A0");
      //Create FFT plans on the last axis
      for (ith = 0; ith < nth; ith++) {
        //cerr << numTransforms[ith] << sep << fftwin[ith] << sep << fftwinOut[ith] << endl;
        exeFFTW(FFTplan[ith] = fftwf_plan_many_dft(rank, &no[2], numTransforms[ith], 
                                           fftwin[ith],   &no[2], no[0] * no[1], 1,
                                           fftwinOut[ith],&no[2], no[0] * no[1], 1,
                                           dir, PLAN_FLAGS));
        //FFTplan[ith] = fftwf_plan_many_dft_r2c(rank, &no[2], n123 / no[2], fftwin[ith], &no[2], no[0] * no[1], 1,
        //                                       fftwin[ith], &no[2], no[0] * no[1], 1, dir, FFTW_MEASURE);
        VERIFY(FFTplan[ith] != NULL);
      }
      //clk.reset("A1");
      //Enter the main compuation part
    #pragma omp parallel default(shared) private(ith)
      {
          ith = omp_get_thread_num();
          //PRV(ith);
          //fftwf_execute(FFTplan[ith]);

          fftwin[ith]         = &fftIn[oTransforms[ith]];
          fftwinOut[ith]      = &fftOut[oTransforms[ith]];
          fftwf_execute_dft(FFTplan[ith],fftwin[ith],fftwinOut[ith]);
      }
      //totRunTime += (clk.reset("A2"));
    } else if (method == TRANSP_FFTW1_FPU) {
      //Now create the FFT plans
      for (ith=0; ith<nth; ith++) {
        fftwin[ith] = fftwf_alloc_complex(n2);
        if (bInPlace) {
          fftwinOut[ith] = fftwin[ith];
        } else {
          fftwinOut[ith] = fftwf_alloc_complex(n2); 
        }
        FFTplan[ith] = fftwf_plan_dft_1d(n2, fftwin[ith], fftwinOut[ith], dir, PLAN_FLAGS); 
      }

      //Do Transpose first, 
      bool bUseSimpleTransp = true;
      if (bUseSimpleTransp) { //first use simple transpose
        for (int i = 0; i < n2; i++) {
          for (int j = 0; j < n1; j++) {
            fftIn[j*n2+i][0] = inCube[i*n1+j];
          }
        }
      } else {
        //Should use more advanced tool like FFTW or MKL
        VERIFY(0);
      }      
      clk.reset("B1");

      //Execute the program
  #pragma omp parallel for default(shared) private(ith)
      for (int i=0; i<n1; i++) {
        ith = omp_get_thread_num();
        //copy the data in
        memcpy(fftwin[ith],fftIn+i*n2,    n2*sizeof(fftwf_complex));
        fftwf_execute(FFTplan[ith]);
        //copy the data out
        memcpy(fftIn+i*n2,fftwinOut[ith], n2*sizeof(fftwf_complex)); 
      }
      
      totRunTime += (clk.reset("B2"));
      //Then do the transpose back
      if (bUseSimpleTransp) { //first use simple transpose
        for (int j = 0; j < n1; j++) {
        for (int i = 0; i < n2; i++) {
            fftOut[i*n1+j][0] = fftIn[j*n2+i][0];
            fftOut[i*n1+j][1] = fftIn[j*n2+i][1];
        }
        }
      } else {
        //Should use more advanced tool like FFTW or MKL
        VERIFY(0);
      }
      //totRunTime += (clk.reset("B2"));
    } else if (method == FFTW2_FPU_MULTH) {
      #ifdef USE_MKL
      //Use FFTW2's multi thread version, here just use one plan, see any difference
      clk.reset("C0");
      VERIFY(0 != fftwf_init_threads()); //MKL does not have this functionality
      fftwf_plan_with_nthreads(nth);
      fftwin[0] = fftIn; fftwinOut[0] = fftOut; 
      ith = 0;
      //create one single plan for the multi-thread purpose.
      FFTplan[0] = fftwf_plan_many_dft(rank, &no[2], n123 / no[2],
                                       fftwin[ith], &no[2], no[0] * no[1], 1,
                                       fftwinOut[ith], &no[2],
                                       no[0] * no[1], 1, dir, PLAN_FLAGS);
      //FFTplan[ith] = fftwf_plan_many_dft_r2c(rank, &no[2], n123 / no[2], fftwin[ith], &no[2], no[0] * no[1], 1,
      //                                       fftwin[ith], &no[2], no[0] * no[1], 1, dir, FFTW_MEASURE);
      VERIFY(FFTplan[0] != NULL);
      scalar *= 1.f / sqrtf(float(no[2]));
      clk.reset("C1");
      //Enter the main compuation part
      fftwf_execute(FFTplan[0]);
      totRunTime += (clk.reset("C2"));
      #else
      VERIFY(0); //we do not have the library.
      #endif
    } else if (method == FFTW2_SSE) {
    } else if (method == FFTW2_FPU2) {
      if (bMakesureAligned) {
        //According to nth, align the memory
        int NS = n1/16; //total number fo segments
        VERIFY(n1%16==0);
        //split the job
        workScheduler1d wrk(NS, nth, Yang::workScheduler1d::chunk);
        int n,o,d;
        for (ith = 0; ith < nth; ith++) {
          wrk.get_nod(ith,&n,&o,&d);
          fftwin[ith]     = &fftIn[o*16];
          fftwinOut[ith]  = &fftOut[o*16]; 
          numTransforms[ith] = n*16;
          oTransforms[ith] = o*16;
          //TRACE(("ith=%d, %d,%d,%d\n",ith,n,o,d));
        }
      } else { //not necessarily aligned.
        //split the job
        int NS = n1;
        workScheduler1d wrk(NS, nth, Yang::workScheduler1d::chunk);
        int n, o, d;
        for (ith = 0; ith < nth; ith++) {
          wrk.get_nod(ith, &n, &o, &d);
          fftwin[ith]         = &fftIn[o];
          fftwinOut[ith]      = &fftOut[o];
          numTransforms[ith]  = n;
          oTransforms[ith] = o;
          if (iter==0) {
            TRACE(("ith=%d, %d,%d,%d\n",ith,n,o,d));
          }
        }
      }
      clk.reset("D0");
      //Create FFT plans on the last axis, but just creat for 1 transform
      for (ith = 0; ith < nth; ith++) {
        //cerr << numTransforms[ith] << sep << fftwin[ith] << sep << fftwinOut[ith] << endl;
        exeFFTW(FFTplan[ith] = fftwf_plan_many_dft(rank, &no[2], 1, 
                                           fftwin[ith],   &no[2], no[0] * no[1], 1,
                                           fftwinOut[ith],&no[2], no[0] * no[1], 1,
                                           dir, PLAN_FLAGS));
        //FFTplan[ith] = fftwf_plan_many_dft_r2c(rank, &no[2], n123 / no[2], fftwin[ith], &no[2], no[0] * no[1], 1,
        //                                       fftwin[ith], &no[2], no[0] * no[1], 1, dir, FFTW_MEASURE);
        VERIFY(FFTplan[ith] != NULL);
      }
      clk.reset("D1");
      //Enter the main compuation part
      int i;
    #pragma omp parallel default(shared) private(ith,i)
      {
          ith = omp_get_thread_num();
          //PRV(ith);
          for (i=0; i<numTransforms[ith];i++) {
            //1 FFT a time
            fftwin[ith]     = &fftIn[oTransforms[ith]+i];
            fftwinOut[ith]  = &fftOut[oTransforms[ith]+i];
            fftwf_execute_dft(FFTplan[ith],fftwin[ith],fftwinOut[ith]); 
          }
      }
      totRunTime += (clk.reset("D2"));
    } else {
      VERIFY(0);
    }
  }
  
  //After the inversion, And save the results
  SEEV(totRunTime);
  cubeData<cmplxf> outCube(shapeOut);
  memcpy(outCube.dataPtr(),(void*)(fftOut[0]),n123*sizeof(cmplxf));
  outCube.dump("out",true);

  //final clean up of the FFTW plans
  fftw_cleanup();
  delete pars;
  return 0;
}

