#include "stdafx.h"
#include "decl_fwd.h"
#include "stackTraceC.h"
#include "cubeData.hpp"
#include "sep_params.h"
#include "misc.h"
#include <complex>
#include "my_timer.h"
#include "myOMP.h"
#include <boost/shared_ptr.hpp>

#include "dmHeaderReader.h"

/**
 * perform Transpose 12, optimized for large file 
 * Usage: Transp12.x <in.H >out.H reshape=2,4 
 */

namespace Yang {

//Transpose 2D array pin to po, pin is of size (n1,n2), pout is of size (n2,n1)
  template<class TOU, class TIN>
  void transp12(TOU* po, const TIN* pin, long long n1, long long n2) {
    long long i1,i2;
    TOU* po2 = po; const TIN* pin2 = pin;
    for (i1=0; i1<n1; i1++) {
      po2 = po+i1*n2; pin2 = pin+i1;
      for (i2=0; i2<n2; i2++) {
        po2[i2] = pin2[i2*n1];
      }
    }
  }

//Use out of core transpose, use file obj dmRsf_oc as input/output medium.
  template<class T>
  int transp12_outofcore(dmRsf_oc<T>& dmOut, dmRsf_oc<T>& dmIn, const hypercube& shpIn12, const hypercube& shpOut12, hypercube& shpIn12_cb, hypercube& shpOut12_cb, cubeData<T>& cbin12, cubeData<T>& cbout12,int* p_nsplit=NULL) {
    FUNC_TRACE_BEG("transp12_outofcore");

    float fFreeMemorySize = getPhysMemorySize()*1e6*0.55f;
    PRV(fFreeMemorySize);
    Yang::timer clk("","clk on transp12_outofcore");

    //split the n1 axis if memory is limited
    int n1 = shpIn12.get_n(1);
    int n2 = shpIn12.get_n(2);
    long long nTot = n1*n2;
    int n_split = nTot*sizeof(T)*2/fFreeMemorySize + 1;
    if (p_nsplit) {
      if (*p_nsplit>=n_split) { //use what the user specified.
        n_split = *p_nsplit;
      } else {
        prt("Warning, user specified n_split=%d will occupy too much memory to cause swapping, reduce n_split to %d\n",*p_nsplit,n_split);
      }
    }

    int n1_split = (n1-1)/n_split+1, n1_blk_last;
    axis ax_cb1 = n1_split;
    axis ax_cb2 = shpIn12.get_axis(2);
    vecAxis axes1; axes1.push_back(ax_cb1); axes1.push_back(ax_cb2);
    vecAxis axes2; axes2.push_back(ax_cb2); axes2.push_back(ax_cb1);
    shpIn12_cb.set_axes(axes1); shpOut12_cb.set_axes(axes2);
    cbin12.alloc(shpIn12_cb); cbout12.alloc(shpOut12_cb); 

    dmIn.reshape(shpIn12); dmOut.reshape(shpOut12);
    dmIn.resetCounter(); dmOut.resetCounter();
    windParam wnd(2,windowAxis());
    wnd[1].set(n2,0,1); wnd[0].j=1;

    PRV4(n_split,n1_split,n1,n2);
    clk.start();
    for (int isp=0; isp<n_split; isp++) {
      prt("Transp12 working on block=%d out of %d ...\n",isp,n_split);
      clk.check("isp1");
      n1_blk_last = min(n1-isp*n1_split,n1_split);
      if (n1_blk_last<=0) break;
      if (n1_blk_last<n1_split) { //redefine the shape of the cubes.
        n1_blk_last = n1-isp*n1_split;
        axes1[0].n = n1_blk_last;
        axes2[1].n = n1_blk_last;
        shpIn12_cb.set_axes(axes1); shpOut12_cb.set_axes(axes2);
        cbin12.alloc(shpIn12_cb); cbout12.alloc(shpOut12_cb); 
      }

      //Then read each block
      wnd[0].n = axes1[0].n; wnd[0].f = isp*n1_split;
      dmIn.grabWindow(&cbin12,wnd);
      clk.check("isp2");
      transp12(cbout12.dataPtr(),cbin12.dataPtr(),axes1[0].n,axes1[1].n);
      dmOut.seq_update(&cbout12);
    }
    clk.check("finished transp");
    FUNC_TRACE_END(0);
  }

//Use out of core transpose, use file obj dmRsf_oc as input/output medium. use srite_window rather than sreed_window.
  template<class T>
  int transp12_outofcore2(dmRsf_oc<T>& dmOut, dmRsf_oc<T>& dmIn, const hypercube& shpIn12, const hypercube& shpOut12, hypercube& shpIn12_cb, hypercube& shpOut12_cb, cubeData<T>& cbin12, cubeData<T>& cbout12,int* p_nsplit=NULL) {
    FUNC_TRACE_BEG("transp12_outofcore2");
    float fFreeMemorySize = getPhysMemorySize()*1e6*0.55f;
    PRV(fFreeMemorySize);
    Yang::timer clk("","clk on transp12_outofcore2");

    //split the n1 axis if memory is limited
    int n1 = shpIn12.get_n(1);
    int n2 = shpIn12.get_n(2);
    long long nTot = n1*n2;
    int n_split = nTot*sizeof(T)*2/fFreeMemorySize + 1;
    if (p_nsplit) {
      if (*p_nsplit>=n_split) { //use what the user specified.
        n_split = *p_nsplit;
      } else {
        prt("Warning, user specified n_split=%d will occupy too much memory to cause swapping, reduce n_split to %d\n",*p_nsplit,n_split);
      }
    }

    //split n2 axis rather than n1 axis
    int n2_split = (n2-1)/n_split+1, n2_blk_last;
    axis ax_cb2 = axis(n2_split);
    axis ax_cb1 = shpIn12.get_axis(1);
    vecAxis axes1; axes1.push_back(ax_cb1); axes1.push_back(ax_cb2);
    vecAxis axes2; axes2.push_back(ax_cb2); axes2.push_back(ax_cb1);
    shpIn12_cb.set_axes(axes1); shpOut12_cb.set_axes(axes2);
    cbin12.alloc(shpIn12_cb); cbout12.alloc(shpOut12_cb); 

    dmIn.reshape(shpIn12); dmOut.reshape(shpOut12);
    dmIn.resetCounter(); dmOut.resetCounter();
    windParam wnd(2,windowAxis());
    wnd[1].set(n1,0,1); wnd[0].j=1;

    PRV4(n_split,n2_split,n1,n2);
    clk.start();
    for (int isp=0; isp<n_split; isp++) {
      prt("Transp12_2 working on block=%d out of %d ...\n",isp,n_split);
      clk.check("isp1");
      n2_blk_last = min(n2-isp*n2_split,n2_split);

      if (n2_blk_last<=0) break;
      if (n2_blk_last<n2_split) { //redefine the shape of the cubes.
        n2_blk_last = n2-isp*n2_split;
        axes1[1].n = n2_blk_last;
        axes2[0].n = n2_blk_last;
        shpIn12_cb.set_axes(axes1); shpOut12_cb.set_axes(axes2);
        cbin12.alloc(shpIn12_cb); cbout12.alloc(shpOut12_cb); 
      }

      //Then read each block
      wnd[0].n = axes2[0].n; wnd[0].f = isp*n2_split;
      dmIn.seq_grab(&cbin12);
      clk.check("isp2");
      transp12(cbout12.dataPtr(),cbin12.dataPtr(),axes1[0].n,axes1[1].n);
      dmOut.updateWindow(&cbout12,wnd);
    }
    clk.check("finished transp2");
    FUNC_TRACE_END(0);
  }

/////Use out of core transpose, use file obj dmRsf_oc as input/output medium. sreed_window and srite_window might be the bottle neck, try not to use that.
  template<class T>
  int transp12_outofcore3(dmRsf_oc<T>& dmOut, dmRsf_oc<T>& dmIn, const hypercube& shpIn12, const hypercube& shpOut12, hypercube& shpIn12_cb, hypercube& shpOut12_cb, cubeData<T>& cbin12, cubeData<T>& cbout12,int* p_nsplit=NULL) {
    FUNC_TRACE_BEG("transp12_outofcore3");

    float fFreeMemorySize = getPhysMemorySize()*1e6*0.55f;
    PRV(fFreeMemorySize);
    Yang::timer clk("","clk on transp12_outofcore");

    //split the n1 axis if memory is limited
    int n1 = shpIn12.get_n(1);
    int n2 = shpIn12.get_n(2);
    long long nTot = n1*n2;
    int n_split = nTot*sizeof(T)*3/fFreeMemorySize + 1;
    if (p_nsplit) {
      if (*p_nsplit>=n_split) { //use what the user specified.
        n_split = *p_nsplit;
      } else {
        prt("Warning, user specified n_split=%d will occupy too much memory to cause swapping, reduce n_split to %d\n",*p_nsplit,n_split);
      }
    }

    int n1_split = (n1-1)/n_split+1, n1_blk_last;
    int n2_split = (n2-1)/n_split+1, n2_blk_last;

    axis ax_cb1 = n1_split;
    axis ax_cb2 = shpIn12.get_axis(2);
    vecAxis axes1; axes1.push_back(ax_cb1); axes1.push_back(ax_cb2);
    vecAxis axes2; axes2.push_back(ax_cb2); axes2.push_back(ax_cb1);
    shpIn12_cb.set_axes(axes1); shpOut12_cb.set_axes(axes2);
    cbin12.alloc(shpIn12_cb); cbout12.alloc(shpOut12_cb); 

    dmIn.reshape(shpIn12); dmOut.reshape(shpOut12);
    dmIn.resetCounter(); dmOut.resetCounter();
    windParam wnd(2,windowAxis());
    wnd[1].set(n2,0,1); wnd[0].j=1;

    PRV4(n_split,n1_split,n1,n2);
    clk.start();
    for (int isp=0; isp<n_split; isp++) {
      prt("Transp12 working on block=%d out of %d ...\n",isp,n_split);
      clk.check("isp1");
      n1_blk_last = min(n1-isp*n1_split,n1_split);
      if (n1_blk_last<=0) break;
      if (n1_blk_last<n1_split) { //redefine the shape of the cubes.
        n1_blk_last = n1-isp*n1_split;
        axes1[0].n = n1_blk_last;
        axes2[1].n = n1_blk_last;
        shpIn12_cb.set_axes(axes1); shpOut12_cb.set_axes(axes2);
        cbin12.alloc(shpIn12_cb); cbout12.alloc(shpOut12_cb); 
      }

      //Then read each block
      wnd[0].n = axes1[0].n; wnd[0].f = isp*n1_split;
      dmIn.grabWindow(&cbin12,wnd);
      clk.check("isp2");
      transp12(cbout12.dataPtr(),cbin12.dataPtr(),axes1[0].n,axes1[1].n);
      dmOut.seq_update(&cbout12);
    }
    clk.check("finished transp12_oc3");
    FUNC_TRACE_END(0);
  } //end function transp12_outofcore3

  //Implement a grab window function by using sequential read and write functions.
  template<class T>
  int grabWindowBySeq_12(cubeData<T>& cb12out, cubeData<T>& cb12tmp, dmRsf_oc<T>&dmIn, const windParam& wnd, int n_split) {
    const axis& ax1 = dmIn.domain().get_axis(1);
    const axis& ax2 = dmIn.domain().get_axis(2);
    int n1=ax1.n, n2=ax2.n;
    int n1_split = (n1-1)/n_split+1, n1_blk_last;
    int n2_split = (n2-1)/n_split+1, n2_blk_last;

    axis ax_cb1 = n1_split;
    axis ax_cb2 = shpIn12.get_axis(2);
    vecAxis axes1; axes1.push_back(ax_cb1); axes1.push_back(ax_cb2);
    vecAxis axes2; axes2.push_back(ax_cb2); axes2.push_back(ax_cb1);
    shpIn12_cb.set_axes(axes1); shpOut12_cb.set_axes(axes2);
    cbin12.alloc(shpIn12_cb); cbout12.alloc(shpOut12_cb); 

    dmIn.reshape(shpIn12); dmOut.reshape(shpOut12);
    dmIn.resetCounter(); dmOut.resetCounter();
    windParam wnd(2,windowAxis());
    wnd[1].set(n2,0,1); wnd[0].j=1;

    PRV4(n_split,n1_split,n1,n2);
    for (int isp=0; isp<n_split; isp++) {
      prt("Transp12 working on block=%d out of %d ...\n",isp,n_split);
      clk.check("isp1");
      n1_blk_last = min(n1-isp*n1_split,n1_split);
      if (n1_blk_last<=0) break;
      if (n1_blk_last<n1_split) { //redefine the shape of the cubes.
        n1_blk_last = n1-isp*n1_split;
        axes1[0].n = n1_blk_last;
        axes2[1].n = n1_blk_last;
        shpIn12_cb.set_axes(axes1); shpOut12_cb.set_axes(axes2);
        cbin12.alloc(shpIn12_cb); cbout12.alloc(shpOut12_cb); 
      }

      //Then read each block
      wnd[0].n = axes1[0].n; wnd[0].f = isp*n1_split;
      dmIn.grabWindow(&cbin12,wnd);
      clk.check("isp2");
      transp12(cbout12.dataPtr(),cbin12.dataPtr(),axes1[0].n,axes1[1].n);
      dmOut.seq_update(&cbout12);
    }
  } //end function

} //end namespace


using namespace Yang;
using namespace std;

int main (int argc, char *argv[])
{
  // TODO: Add application code here
  initCallStackTrace();
  init_omp(0);
  //have to init par
  sep_params* pars = new sep_params(argc,argv);
  sepHeaderReader inHeader("in",dm::IN,"",true); //read the header information

  int n_r_read;
  int nmax = 7;
  stdVecInt ndef(nmax,-1);
  int* reshape  = pars->get_ints("reshape",nmax,&ndef[0],&n_r_read);
  bool verb     = pars->get_bool("verb",true);
  int nsplit = pars->get_int("nsplit",-1);
  int* p_nsplit = &nsplit;
  if (nsplit==-1) {
    p_nsplit = NULL;
  };

  if (verb) {
    PRV(nsplit);
  }
  int ndim_default = 2;
  if (n_r_read==0) { //no need to reshape, use default
    n_r_read = ndim_default;
    reshape = new int[ndim_default];
    for (int i=1;i<=ndim_default;i++) {
      reshape[i-1] = i;
    }
  } else {
    VERIFY(n_r_read ==2);
  }
  int ndim = reshape[n_r_read-1];
  hypercube inShp,outShp;
  inHeader.getShape(inShp,ndim);

  int esize = inHeader.getElemSize();

  if (verb) {
    PRV2(esize,ndim);
  }

  dmRsf_oc<float>   *pdmI_4b=NULL,*pdmO_4b=NULL;
  dmRsf_oc<cmplxf>  *pdmI_8b=NULL,*pdmO_8b=NULL;
  dmRsf_oc<byte>    *pdmI_1b=NULL,*pdmO_1b=NULL;

  cubeData<float>   *pcbI_4b=NULL,*pcbO_4b=NULL;
  cubeData<cmplxf>  *pcbI_8b=NULL,*pcbO_8b=NULL;
  cubeData<byte>    *pcbI_1b=NULL,*pcbO_1b=NULL;

  vecAxis axes_in12,axes_out12,axes_out,axes_t;
  vecAxis axes = inShp.get_axes();

  hypercube shpIn12,shpOut12,shpIn12_cb,shpOut12_cb;
  int id=0; long long n123ax;
  for (id=0;id<n_r_read;id++) {
    //lump the axes together
    axis ax; axes_t.clear();
    if (id==0) {
      ax = axes[0]; n123ax = ax.n; axes_t.push_back(ax);
      for (int jd=1;jd<=reshape[id]-1;jd++) {
        n123ax *= axes[jd].n; axes_t.push_back(axes[jd]);
      }
    } else {
      ax = axes[reshape[id-1]]; n123ax = ax.n; axes_t.push_back(ax);
      for (int jd=reshape[id-1]+1;jd<=reshape[id]-1;jd++) {
        n123ax *= axes[jd].n; axes_t.push_back(axes[jd]);
      }
    }

    VERIFY(n123ax<2e9);
    ax.n = n123ax;
    axes_in12.push_back(ax);
    axes_out12.insert(axes_out12.begin(),ax);
    axes_out.insert(axes_out.begin(),axes_t.begin(),axes_t.end());
  }

  shpIn12.set_axes(axes_in12); shpOut12.set_axes(axes_out12); 
  outShp.set_axes(axes_out);

  inShp.info("inShp");outShp.info("outShp");
  shpIn12.info("shpIn12"); shpOut12.info("shpOut12");

  if (esize==1) {
    pdmI_1b = new dmRsf_oc<byte>("in",dm::IN,"input",true);
    pdmO_1b = new dmRsf_oc<byte>("out",dm::OUT,"otuput",true);
    pcbI_1b = new cubeData<byte>(); pcbO_1b = new cubeData<byte>();
    pdmI_1b->init(NULL,ndim); pdmO_1b->init(&outShp,ndim);
    transp12_outofcore(*pdmO_1b, *pdmI_1b, shpIn12, shpOut12, shpIn12_cb, shpOut12_cb, *pcbI_1b, *pcbO_1b,p_nsplit);
  } else if (esize==4) {
    pdmI_4b = new dmRsf_oc<float>("in",dm::IN,"input",true);
    pdmO_4b = new dmRsf_oc<float>("out",dm::OUT,"otuput",true);
    pcbI_4b = new cubeData<float>(); pcbO_4b = new cubeData<float>();
    LOG("F2");
    pdmI_4b->init(NULL,ndim); pdmO_4b->init(&outShp,ndim);    
    LOG("F3");
    transp12_outofcore(*pdmO_4b, *pdmI_4b, shpIn12, shpOut12, shpIn12_cb, shpOut12_cb, *pcbI_4b, *pcbO_4b,p_nsplit);
    LOG("F4");
  } else if (esize==8) {
    pdmI_8b = new dmRsf_oc<cmplxf>("in",dm::IN,"input",true);
    pdmO_8b = new dmRsf_oc<cmplxf>("out",dm::OUT,"otuput",true);
    pcbI_8b = new cubeData<cmplxf>(); pcbO_8b = new cubeData<cmplxf>();
    pdmI_8b->init(NULL,ndim); pdmO_8b->init(&outShp,ndim);    
    transp12_outofcore(*pdmO_8b, *pdmI_8b, shpIn12, shpOut12, shpIn12_cb, shpOut12_cb, *pcbI_8b, *pcbO_8b,p_nsplit);
  } else {
    VERIFY(0);
  }
}

