#include "stdafx.h"
#include "decl_fwd.h"
#include "stackTraceC.h"
#include "cubeData.hpp"
#include "sep_params.h"
#include "misc.h"
#include <fftw3.h>
#include <complex>
#include "my_timer.h"
#include "myOMP.h"
#include <boost/shared_array.hpp>
#include <boost/shared_ptr.hpp>

#include "fft3d.h"

/**
 * class that performs 3-D fourier transform
 */

#ifdef INPUT_TYPE
  //#define INPUT_TYPE float
#else
  #error "specify macro INPUT_TYPE to float or cmplxf
#endif

using namespace Yang;
using namespace std;

void fft3d::readParam(const param_func* pars, const hypercube& inShape) {
  this->inShape_ = inShape;

  sign_[0] = pars->get_int("sign1",0); VERIFY(sign_[0]==-1 || sign_[0]==1 || sign_[0]==0);
  sign_[1] = pars->get_int("sign2",0); VERIFY(sign_[1]==-1 || sign_[1]==1 || sign_[1]==0);
  sign_[2] = pars->get_int("sign3",0); VERIFY(sign_[2]==-1 || sign_[2]==1 || sign_[2]==0);
  ndim_ = inShape.get_ndim();
  n_[1]=n_[2]=n_[0]=1;
  n_[0] = inShape.get_n(1); n_[1] = inShape.get_n(2);
  if (ndim_ > 1) {
    n_[1] = inShape.get_n(2);
  }
  if (ndim_ > 2) {
    n_[2] = inShape.get_n(3);
  }

  //The output dimensions info
  myMemcpy(no_,n_,3);
  if (sign_[0]!=0) {
    no_[0] = pars->get_int("n1",n_[0]); VERIFY(no_[0]>=n_[0])
  }; 
  if (sign_[1]!=0) {
    no_[1] = pars->get_int("n2",n_[1]); VERIFY(no_[1]>=n_[1])
  }; 
  if (sign_[2]!=0) {
    no_[2] = pars->get_int("n3",n_[2]); VERIFY(no_[2]>=n_[2])
  };

  vecAxis axes = inShape_.get_axes(); axis ax;
  for (int i=0; i<min(3,ndim_); i++) {
    if (sign_[i]!=0) { //change axis 1
      get_axis_inv(axes[i], no_[i], axes[i]);
    }
  }
  outShape_.set_axes(axes);

  long long n123 = outShape_.get_n123();
  //each cmplxf cube cannot extend 2GB. Remember we have three such cubes, cb_read,cbin,cbout,cb_write. If exceed, the it cause memory swapping, and on cees-cluster nodes, swapping would happen for size 4GB. Therefore to be conservative, use 2GB as limit.
  if(n123*4 > 1024*1024*1024) {
    prt("The size is too big in fft3d class, which might cause memory swapping!\n");
  }
  return;
}

fft3d::~fft3d() {
  int n = pOutCubeVecs_.size();
  for (int i=0; i<n; i++) {
    delete pOutCubeVecs_[i];
    fftwf_free(fftwin2d_[i]);
    fftwf_free(fftwin_[i]);
    fftwf_destroy_plan(FFTplan_[0][i]);
    fftwf_destroy_plan(FFTplan_[1][i]);
    fftwf_destroy_plan(FFTplan_[2][i]);
  }
}

void Yang::fft3d::before_trans() {
  int nth = g_NUM_THREADS;
  pOutCubeVecs_.resize(nth);
  hypercube shpOutcube = hypercube(no_[0],no_[1],no_[2]);
  int i,ith;
  
  hypercube shpIncube = hypercube(n_[0],n_[1],n_[2]);
  inCube_.alloc(shpIncube);

  for (i=0; i<nth; i++) {
    pOutCubeVecs_[i] = new cubeData<cmplxf>(shpOutcube);
  }

  int rank=1; int dir;

  //create the fftw plan
  long long n123 = no_[0]*no_[1]*no_[2];

  fftwin_.resize(nth); fftwin2d_.resize(nth);
  for (ith=0;ith<nth;ith++) {
    fftwin_[ith]  = (fftwf_complex*)fftwf_malloc(sizeof(fftwf_complex)*n123);           memset(fftwin_[ith],0,n123*sizeof(fftwf_complex));
    fftwin2d_[ith]= (fftwf_complex*)fftwf_malloc(sizeof(fftwf_complex)*no_[0]*no_[1]);  memset(fftwin2d_[ith],0,no_[0]*no_[1]*sizeof(fftwf_complex));
  }

  scalar_ = 1.0f;
  //along axis 1
  FFTplan_[0].resize(nth); FFTplan_[1].resize(nth); FFTplan_[2].resize(nth);
  if (sign_[0]!=0) {
    dir = (sign_[0]==-1)? FFTW_FORWARD:FFTW_BACKWARD;
    for (ith=0;ith<nth;ith++) {
      FFTplan_[0][ith] = fftwf_plan_many_dft(rank,&no_[0],n123/no_[0], fftwin_[ith], &no_[0],1          ,no_[0], fftwin_[ith],&no_[0],1,no_[0], dir, FFTW_MEASURE);
      VERIFY( FFTplan_[0][ith]!=NULL);
    }
    scalar_ *= 1.f/sqrtf(float(no_[0]));
  }
  //along axis 3
  if (sign_[2]!=0) {
    dir = (sign_[2]==-1)? FFTW_FORWARD:FFTW_BACKWARD;
    for (ith=0;ith<nth;ith++) {
      FFTplan_[2][ith] = fftwf_plan_many_dft(rank,&no_[2],n123/no_[2], fftwin_[ith], &no_[2],no_[0]*no_[1],1,     fftwin_[ith],&no_[2],no_[0]*no_[1],1, dir, FFTW_MEASURE);
      VERIFY( FFTplan_[2][ith]!=NULL);
    }
    scalar_ *= 1.f/sqrtf(float(no_[2]));
  }
  //along axis 2 cannot be done at once. need to do each 3D slice by slice.
  if (sign_[1]!=0) {
    dir = (sign_[1]==-1)? FFTW_FORWARD:FFTW_BACKWARD;
    for (ith=0;ith<nth;ith++) {
      FFTplan_[1][ith] = fftwf_plan_many_dft(rank,&no_[1],no_[0]     , fftwin2d_[ith], &no_[1],no_[0]      ,1,     fftwin2d_[ith],&no_[1],no_[0]      ,1, dir, FFTW_MEASURE);
      VERIFY( FFTplan_[1][ith]!=NULL);
    }
    scalar_ *= 1.f/sqrtf(float(no_[1]));
  }
}

void Yang::fft3d::transform(cubeData<cmplxf>& cbin, cubeData<cmplxf>& cbout) {
  //Do the fft for each 3d cube
  hypercube shpOutcube = hypercube(no_[0],no_[1],no_[2]);
  int ndim = this->ndim_;
  int nth = g_NUM_THREADS, ith;
  int nRead = 1; if (ndim>3) {
    nRead = this->inShape_.get_n23(4,ndim);
  }
  long long n123 = no_[0]*no_[1]*no_[2];
  VERIFY(n123 < 2e9);

  windParam wnd(3,windowAxis());
  wnd[0].set(n_[0],0,1); wnd[1].set(n_[1],0,1); wnd[2].set(n_[2],0,1);

  //Enter the main compuation part
  int iRead = 0;
  cbin.resetCounter(); cbout.resetCounter();

#pragma omp parallel default(shared) private(iRead,ith) 
  {
    ith = omp_get_thread_num();
    for (iRead=0; iRead<nRead; iRead+=nth) {
      int mRead = std::min(nth,nRead-iRead);
      //Only the thread 0 does the IO
      if (ith==0) {
        for (int i = 0; i<mRead; i++) {
          cbin.seq_grab(&inCube_);
          pOutCubeVecs_[i]->zero();
          pOutCubeVecs_[i]->updateWindow(&inCube_,wnd);
          //Then copy to fftwin_ array
          memcpy(fftwin_[i],pOutCubeVecs_[i]->dataPtr(),sizeof(cmplxf)*n123);
        }
      }

      #pragma omp barrier //set a barrier
      //Every thread that gets the data does the work
      if (ith < mRead) {
        //Do first axis
        if (sign_[0]!=0) {
          fftwf_execute(FFTplan_[0][ith]);
        }
        //axis 3
        if (sign_[2]!=0) {
          fftwf_execute(FFTplan_[2][ith]);
        }
        //axis 2
        if (sign_[1]!=0) {
          int n12 = no_[0]*no_[1];
          for (int i3=0; i3<no_[2];i3++) {
            myMemcpy(fftwin2d_[ith],fftwin_[ith]+i3*n12,n12);
            fftwf_execute(FFTplan_[1][ith]);
            myMemcpy(fftwin_[ith]+i3*n12,fftwin2d_[ith],n12);
          }
        }
        //load the data into the arrays
        memcpy(pOutCubeVecs_[ith]->dataPtr(),fftwin_[ith],sizeof(cmplxf)*n123);
        pOutCubeVecs_[ith]->scale(cmplxf(scalar_,0.f));
      }
      //let all thread finish the computation
      #pragma omp barrier 

      //do the write out
      if (ith == 0) {
        for (int i = 0; i<mRead; i++) {
          cbout.seq_update(pOutCubeVecs_[i]);
        }
      }
    } //end for iRead
  } //end omp parallel
}

//Transpose two arrays, pi is n1 by n2, po is n2 by n1
template<class TOU, class TIN>
void transp12(TOU* po, const TIN* pin, long long n1, long long n2) {
  long long i1,i2;
  TOU* po2 = po; const TIN* pin2 = pin;
  for (i1=0; i1<n1; i1++) {
    po2 = po+i1*n2; pin2 = pin+i1;
    for (i2=0; i2<n2; i2++) {
      po2[i2] = pin2[i2*n1];
    }
  }
}

//Doing 3D fourier transform on axis 3,4,5 on a cube of (1,2,3,4,5), using fft3d class
int main (int argc, char *argv[])
{
  // TODO: Add application code here
  initCallStackTrace();
  init_omp(0);
  //have to init par
  sep_params* pars = new sep_params(argc,argv);  
  int nth_par = pars->get_int("nth",8); //default is 8 for cees-cluster, should be 15 for cees-rcf
  if (nth_par != -1) { //we read something from nth
    VERIFY(nth_par>0);
    change_omp(nth_par);
  }
  PRV(g_NUM_THREADS);

  fft3d trans_fft3d;
  bool bNotDummy = true;
  dmRsf_oc<INPUT_TYPE> dmIn("in",Yang::dm::IN,"input",bNotDummy);  
  dmRsf_oc<cmplxf> dmOut("out",Yang::dm::OUT,"output",bNotDummy);
  
  //int ndim = 5;
  dmIn.init(NULL,5); //obtains its shape
  hypercube shpIn = dmIn.domain();
  vector<axis>  axesIn = shpIn.get_axes();
  vector<int>   ns = shpIn.get_ns();
  axis ax1 = axesIn[0], ax2 = axesIn[1];

  cubeData<INPUT_TYPE>  cb_read;
  cubeData<cmplxf> cbin,cbout,cb_write;

  int n1,n2, n_block, n1_t;
  n1 = ns[0]; n2 = ns[1]; 
  n_block = pars->get_int("n_block",n1); //read a whole chunk of (n1,1,n3,n4,n5 packet)
  VERIFY(n1%n_block==0); //currently cannot handle other cases.

  hypercube inShp_incore = shpIn.return_compatible(3,5);
  vecAxis axes_inshp_ic = inShp_incore.get_axes();
  
  hypercube shp_cbin,shp_read,shp_cbout,shp_write;
  axis ax_blk(n_block,0.f,1.0f);

  axes_inshp_ic.push_back(ax_blk);
  shp_cbin.set_axes(axes_inshp_ic); cbin.alloc(shp_cbin);
  axes_inshp_ic.pop_back(); axes_inshp_ic.insert(axes_inshp_ic.begin(),ax_blk);
  shp_read.set_axes(axes_inshp_ic); cb_read.alloc(shp_read);

  shp_cbin.info("shp cbin");    shp_read.info("shp read");

  trans_fft3d.readParam(pars,shp_cbin);
  trans_fft3d.before_trans();
  
  LOG("D2");
  //Then build a temporary arrays for a cube of (3,4,5,nth*(1*2/nth)) for applying fft on 3,4,5 
  windParam wnd(5,windowAxis()),wnd_out(5,windowAxis());
  vecAxis axes2 = trans_fft3d.outShape().get_axes();
  axes2.pop_back(); axes2.insert(axes2.begin(),ax2); axes2.insert(axes2.begin(),ax1);
  hypercube shpOut(axes2);  dmOut.init(&shpOut);
  shpOut.info("shpOut total");

  vector<int> ns_out = shpOut.get_ns();
  wnd[2].set(ns[2],0,1); wnd[3].set(ns[3],0,1); wnd[4].set(ns[4],0,1);
  wnd_out[2].set(ns_out[2],0,1); wnd_out[3].set(ns_out[3],0,1); wnd_out[4].set(ns_out[4],0,1);
  
  vecAxis axes_out = trans_fft3d.outShape().get_axes();
  shp_cbout.set_axes(axes_out); cbout.alloc(shp_cbout);
  axes_out.pop_back(); axes_out.insert(axes_out.begin(),ax_blk);
  shp_write.set_axes(axes_out); cb_write.alloc(shp_write);
  shp_cbout.info("shp cbout");  shp_write.info("shp write");

  Yang::timer clk_1("clk created", "No.1"); clk_1.start();
  int n2in,n2out;
  n2in = ns[2]*ns[3]*ns[4]; n2out = ns_out[2]*ns_out[3]*ns_out[4];
  int i1,i2;
  for (i2=0;i2<n2;i2++) {
    prt("processing i2=%d out of %d \n", i2,n2);
    for (i1=0;i1<n1;i1+=n_block) {
      n1_t = std::min(n_block,n1-i1);
      wnd[0].set(n1_t,i1,1); wnd[1].set(1,i2,1);
      clk_1.reset("start read chunk file"); clk_1.go();
      dmIn.grabWindow(&cb_read,wnd);
      clk_1.reset("finish read chunk file");clk_1.go();
      //transpose between cbin and cb_read
      transp12(cbin.dataPtr(),cb_read.dataPtr(),n_block,n2in);
      clk_1.reset("finish input transpose");clk_1.go();
      trans_fft3d.transform(cbin,cbout);
      clk_1.reset("finish block transform");clk_1.go();
      transp12(cb_write.dataPtr(),cbout.dataPtr(),n2out,n_block);
      clk_1.reset("finish output transpose");clk_1.go();
      wnd_out[0].set(n1_t,i1,1); wnd_out[1].set(1,i2,1);
      dmOut.updateWindow(&cb_write,wnd_out);
      clk_1.reset("finish write chunk file");clk_1.go();
    }
  }
}

