#include "stdafx.h"
#include "cubeData.hpp"
#include "op.h"
#include "cblas_RSF.h"
#include "istc_solver.h"

#include "algo.h"
#include "cubeAlgo2.h"

using namespace Yang;

//Generate vec from 0.0-1.0, uniform distribution
void random_vec (float* x, long long n) {
  VERIFY(RAND_MAX > 32000);
  for (long long i=0; i<n; i++) {
    x[i] = float(rand() % RAND_MAX) / float(RAND_MAX) ;
  }
}

int Yang::istcSolver (op& Foper,   //linear operator
                  float* x,           // estimated model
                  long long nx,
                  const float* dat,   // data
                  long long nd,
                  int nlIter,          // non-linear iter
                  int nIter, //number of linear iterations
                  const int* region,
                  long long nr,
                  const float* start,
                  long long ns,
                  const float* pEigenval,
                  const float* x0,
                  bool bUseInputMsk,  //model space msk
                  float* msk_model,
                  bool verb) {
  bool b_x0 = x0!=NULL;
  long long i,j;
  
  if (b_x0) {
    for (i=0; i<nx; i++) { x[i]=x0[i]; }
  } else {
    LOG("Randomize x in istc_solver");
    random_vec(x,nx);
  }

  valarray<float>  reg(ns);

  const hypercube& shpMod = Foper.ModShape();

  //Begin by cooling method
  int region_max = *(std::max_element(region,region+nr));

  hypercube tau_shp(region_max,nlIter);
  cubeDataF tau(tau_shp);
  valarray<float> gg(nd),rr(nd),g(nx);

  random_vec(va_ptr(g),g.size());

  float sc = 0.f, rdot;

  bool bHaveEigenValueAlready = (pEigenval!=NULL);
  if (bHaveEigenValueAlready) {
    sc = *pEigenval;
  } else {
    //First find the largest eigen-value of the Hessian matrix through power-method, just do it once.
    int nTrialPowerMethod=15;
    for (int iter=1; iter<=nTrialPowerMethod; iter++) {
      Foper.oper(false,false,nx,nd,x,va_ptr(gg)); //fwd
      Foper.oper(true,false,nx,nd,va_ptr(g),va_ptr(gg)); //adj
      rdot = cblas_snrm2(nx,va_ptr(g),1);
      sc = sqrtf(rdot);
      VERIFY(sc > epsilon);
      g = g*(1.0f/sc);
      if (verb) prt("Check this, %d, %f, %f \n", iter, sc, rdot);
      for (i=0; i<nx; i++) {x[i]=g[i];}
    }
  }

  if (b_x0) { 
    for (i=0; i<nx; i++) {x[i]=x0[i];}
  } else {
    for (i=0; i<nx; i++) {x[i]=0.f;}
  }

  //Make it 1.0/eigenvalue
  sc = 1.0f/sc;
  tau.info("tau in istc_solver");

  tau.zero();
  float* dat_cc = const_cast<float *>(dat);
  Foper.oper(true,false,nx,nd,va_ptr(g),dat_cc); //compute data back projection gradient.

  float rat = 1e4 / cblas_snrm2(nd,dat_cc,1);
  dumpStdArray(g,"g.H",true,&shpMod);

  g = std::abs(g)*sc;
  LOG("DYING SOME1");

  std::valarray<float> tmp; float val;
  VERIFY(nlIter>0);

  float fPctTarget = 0.06f; //Target is keep 5% of coefficients
  PRV(region_max);

  bool bHasModelMsk = (NULL!=msk_model);
  if (bUseInputMsk) {
    VERIFY(bHasModelMsk);
  } else { //generate quantile criterion to do the thresholding.
    tmp.resize(nx); tmp=std::abs(g);
    arrayStat("array 'g' in istc solver",va_ptr(g),nx);
    if (start[0]<0.f) {
      //val=.005f;
      val = fPctTarget / powf(1.8f,float(nlIter-1)); //
      for (i=0; i<nlIter; i++) {
        for (j=0; j<region_max; j++) {
          tau(i,j)=quantile(int((1.f-val)*tmp.size()),va_ptr(tmp),tmp.size());
          tau(i,j) *= sqrtf(sc);
        }
        PRV(nx); TRACE(("CHECK QUANTILE, val=%f, val2=%d, tau0=%f, \n",val,lroundf((1.f-val)*nx),tau(i,0)));
        val *= 1.8f;
      }
      if (bHasModelMsk) { //dump the quantile msk to msk_model array
        for (i=0; i<nx; i++) {
          msk_model[i] = float(tmp[i]>tau(nlIter-1,0));
        }
      }
    } else {
      int ncount;
      for (j=0;j<ns;j++) {
        ncount=0;
        val=start[j]/100.f;
    
        for (i=0;i<nx;i++) {
          if(region[i]==j+1) ncount++;
        }
    
        LOG("DYING 3");
        tmp.resize(ncount);
        TRACE(("FOUND %d %d \n",j,ncount));
      
        ncount=0;
        for (i=0; i<nx; i++) {
          if(region[i]==j+1) {
            ncount++;
            tmp[ncount-1]= g[i];
          }
        }
      
        for (i=0; i<nlIter-1; i++) {
          tau(i,j)=quantile(floor((1.f-val)*tmp.size()),va_ptr(tmp),tmp.size());
          val=std::min(val*1.8f,0.99999f);
          if(verb) PRV4(j,i,val,tau(i,j));
        } //end for i
      } //end for j
    } //end if
  } //end fi use_input_msk
  
  
  //Loop over cooling steps
  int initer=0, iter,nkill;
  for (initer=1; initer<=nlIter; initer++) {
    //if(verb) {prt("nlIter=%d threshold=%f",initer,tau(:,initer));}
    
    //Power method
    iter=1;
    float fold=1e6, fnew=fold-5.0f;
    while (iter <=nIter && (fold-fnew >0.05f) && fnew >0.99f) {
      fold = fnew;
      Foper.oper(false,false,nx,nd,x,va_ptr(gg));
      for (i=0; i<nd; i++) { gg[i] = dat[i]-gg[i];}

      // gg=dat-gg current residual
      fnew = cblas_snrm2(nd,va_ptr(gg),1) * rat;
      Foper.oper(true,false,nx,nd,va_ptr(g),va_ptr(gg));
      
      for (i=0; i<nx; i++) { g[i] = x[i]+g[i]*sc;}  //g=x+g*sc 

      nkill = 0;
      if (bUseInputMsk) { //use InputMsk to do thresholding
        for (i=0; i<nx; i++) { x[i] = g[i]*msk_model[i];}
      } else {
        for (i = 0;  i < nx; i++) {
          if(std::abs(g[i])>tau(initer-1,region[i]-1)) {
            x[i]=g[i];
          } else {
            nkill++; x[i]=0.f;
          }
        }
      }
      if(verb) prt("iter=%d, magnitude=%f , %f , %f \n", iter, fnew,float(nkill)/float(nx),(fold-fnew)/10.f);
      iter++;
    } //end while
  } //end for initer
  return 0;
} //end function

